"""emv.py
"""

# Standard library imports

# Third party imports

# Local application imports
from pcsc import card
from emv import commands as command
from emv import records

def SELECT(card_service: card._CardService, aid: str, trace=None):
    return card.send_apdu(card_service, command.SELECT(aid), trace=trace)

def GPO(card_service: card._CardService, pdol: str, trace=None):
    return card.send_apdu(card_service, command.GPO(pdol), trace=trace)

def READ_RECORD(card_service: card._CardService, sfi: int, record:int, trace=None):
    return card.send_apdu(card_service, command.READ_RECORD(sfi, record), trace=trace)

def READ_RECORDs(card_service: card._CardService, AFL: str, trace=None):
    output = []

    r, err = records.decode_afl(AFL)
    if err is not None:
        return []
    else:
        for (sfi, first, last, _) in r:
            for n in range(first, last+1):
                r, err = READ_RECORD(card_service, sfi, n, trace=trace)
                output.append((r, err))
                if err != '':
                    return output
        return output

def GET_DATA(card_service: card._CardService, tag: str, trace=None):
    return card.send_apdu(card_service, command.GET_DATA(tag), trace=trace)



"""card.py
"""

# Standard library imports

# Third party imports
from smartcard import CardType as _CardType
from smartcard.CardRequest import CardRequest as _CardRequest
from smartcard.Exceptions import CardRequestTimeoutException as _CardRequestTimeoutException
from smartcard.scard import \
    SCARD_SHARE_SHARED as _SCARD_SHARE_SHARED,\
    SCARD_SHARE_EXCLUSIVE as _SCARD_SHARE_EXCLUSIVE,\
    SCARD_SHARE_DIRECT as _SCARD_SHARE_DIRECT,\
    SCARD_LEAVE_CARD as _SCARD_LEAVE_CARD,\
    SCARD_RESET_CARD as _SCARD_RESET_CARD,\
    SCARD_UNPOWER_CARD as _SCARD_UNPOWER_CARD,\
    SCARD_EJECT_CARD as _SCARD_EJECT_CARD
from smartcard.CardConnection import CardConnection as _CardConnection
from smartcard.CardService import CardService as _CardService
from smartcard.CardMonitoring import \
    CardMonitor as _CardMonitor,\
    CardObserver as _CardObserver
from smartcard.System import readers as _readers

# Local application imports
from common.intlist import to_hstr as _to_hstr
from common.hstr import \
    to_intlist as _to_intlist,\
    minimum as _min,\
    clean as _clean


def connect(atr=None,
            atr_mask=None,
            new_card_only=False,
            readers=None,
            timeout=None,
            protocol=None,
            mode='SCARD_SHARE_EXCLUSIVE',
            disposition='SCARD_LEAVE_CARD',
            trace=None) -> _CardService:
    """connect()
    """
    if atr is None:
        # Requesting any card
        _trace(trace, "Requesting any card")
        card_type = _CardType.AnyCardType()
    elif atr_mask is None:
        # Requesting a card with a specific ATR
        _trace(trace, "Requesting a card with a specific ATR")
        card_type = _CardType.ATRCardType(_to_intlist(atr))
    else:
        # Requesting a card with a masked ATR
        _trace(trace, "Requesting a card with a masked ATR")
        card_type = _CardType.ATRCardType(
            _to_intlist(atr, command_name='card.connect()', hstr_name='atr'),
            _to_intlist(atr_mask, command_name='card.connect()', hstr_name='atr_mask'))

    card_request = _CardRequest(
        newcardonly=new_card_only,
        readers=readers,
        cardType=card_type,
        timeout=timeout)

    if protocol is not None:
        _trace(trace, F"Requested protocol: {protocol}")
        protocol = {'SCARD_PROTOCOL_T0': _CardConnection.T0_protocol,
                    'SCARD_PROTOCOL_T1': _CardConnection.T1_protocol,
                    'SCARD_PROTOCOL_T0_OR_T1':
                    _CardConnection.T0_protocol |
                    _CardConnection.T1_protocol,
                    'SCARD_PROTOCOL_RAW': _CardConnection.RAW_protocol}[protocol]

    _trace(trace, F"Requested mode: {mode}")
    mode = {'SCARD_SHARE_SHARED': _SCARD_SHARE_SHARED,
            'SCARD_SHARE_EXCLUSIVE': _SCARD_SHARE_EXCLUSIVE,
            'SCARD_SHARE_DIRECT': _SCARD_SHARE_DIRECT}[mode]

    _trace(trace, F"Requested disposition: {disposition}")
    disposition = {'SCARD_LEAVE_CARD': _SCARD_LEAVE_CARD,
                   'SCARD_RESET_CARD': _SCARD_RESET_CARD,
                   'SCARD_UNPOWER_CARD': _SCARD_UNPOWER_CARD,
                   'SCARD_EJECT_CARD': _SCARD_EJECT_CARD}[disposition]

    try:
        _trace(trace, F"Card request will timeout after {timeout}s")
        card_service = card_request.waitforcard()
        card_service.connection.connect(
            protocol=protocol,
            mode=mode,
            disposition=disposition)
    except _CardRequestTimeoutException as err:
        return '', err
    else:
        _trace(
            trace, F"Card {get_atr(card_service)} detected on {get_reader(card_service)}")
        return card_service, ''


def get_reader(card_service: _CardService) -> str:
    """get_Reader()
    """
    return card_service.connection.getReader()


def list_readers():
    """list_readers()
    """
    return _readers()


def get_atr(card_service: _CardService) -> str:
    """getATR()
    """
    return _to_hstr(card_service.connection.getATR())


def get_protocol(card_service: _CardService) -> str:
    """getProtocol()
    """
    return {_CardConnection.T0_protocol: 'T=0',
            _CardConnection.T1_protocol: 'T=1',
            _CardConnection.T0_protocol & _CardConnection.T1_protocol: 'T=0 and T=1',
            _CardConnection.RAW_protocol: 'RAW'}[card_service.connection.getProtocol()]


def transmit(card_service: _CardService, hstr: str, trace=None):
    """transmit()
    """
    _trace(trace, F"=>card: {hstr}")

    data, sw1, sw2 = card_service.connection.transmit(
        _to_intlist(hstr), card_service.connection.getProtocol())
    data, sw1, sw2 = _to_hstr(data), F"{sw1:02X}", F"{sw2:02X}"

    if data == '':
        _trace(trace, F"<=card: {sw1+sw2}")
    else:
        _trace(trace, F"<=card: {data} {sw1+sw2}")

    return [data, sw1, sw2], ''


def send_apdu(card_service: _CardService, apdu: dict, trace=None):
    """send_apdu()
    """
    # Extracting APDU header, data, length field from apdu dictionary
    if 'header' in apdu:
        header = _clean(apdu['header'])
    else:
        CLA = _clean(apdu['CLA'])
        INS = _clean(apdu['INS'])
        P1 = _clean(apdu['P1'])
        P2 = _clean(apdu['P2'])
        header = CLA + INS + P1 + P2

    data = apdu.get('data', None)
    Le = apdu.get('Le', None)

    protocol = get_protocol(card_service)
    if protocol not in ['T=0', 'T=1']:
        return ['', '', ''], 'Unsupported protocol: ' + protocol

    if data is None:
        # Case 1
        if Le is None:
            if protocol == 'T=0':
                return _send_apdu_T0_case_1(card_service, header, trace)
            if protocol == 'T=1':
                return transmit(card_service, header, trace)
        # Case 2S
        Le = _clean(Le, 'send_apdu()', 'Le')
        if len(Le) == 2:
            if protocol == 'T=0':
                return _send_apdu_T0_case_2s(card_service, header, Le, trace)
            if protocol == 'T=1':
                return transmit(card_service, header + Le, trace)
        # Case 2E
        if len(Le) == 6:
            if protocol == 'T=0':
                return ['', '', ''], 'Unsupported case 2E with protocol T=0'
            if protocol == 'T=1':
                return transmit(card_service, header + Le, trace)

    data = _clean(data, 'send_apdu()', 'data')
    Lc = len(data)//2
    if Le is None:
        # Case 3S
        if Lc < 256:
            Lc = F"{Lc:02X}"
            if protocol == 'T=0':
                return _send_apdu_T0_case_3s(card_service, header, Lc, data, trace)
            if protocol == 'T=1':
                return transmit(card_service, header + Lc + data, trace)
        # Case 3E
        if Lc < 65536:
            Lc = F"00{Lc:04X}"
            if protocol == 'T=0':
                return ['', '', ''], 'Unsupported case 3E with protocol T=0'
            if protocol == 'T=1':
                return transmit(card_service, header + Lc + data, trace)
        # Data field too long
        else:
            return ['', '', ''], F"Data field too long: {Lc:X}"

    # Case 4S
    if Lc < 256 and len(Le) == 2:
        if protocol == 'T=0':
            return _send_apdu_T0_case_4s(card_service, header, F"{Lc:02X}", data, Le, trace)
        if protocol == 'T=1':
            return transmit(card_service, header + F"{Lc:02X}" + data + Le, trace)
    # Case 4E
    if Lc < 65536 and len(Le) == 6:
        if protocol == 'T=0':
            return ['', '', ''], 'Unsupported case 4E with protocol T=0'
        if protocol == 'T=1':
            return transmit(card_service, header + F"00{Lc:04X}" + data + Le, trace)

    return ['', '', ''], F'Unsupported case with short and extended lengths: Lc = {Lc:X}, Le = {Le}'


def _send_apdu_T0_case_1(card_service: _CardService, header: str, trace=None):
    """send_apdu_T0_case_1():
    """
    _trace(trace, "Case 1 Command (T=0)")
    return transmit(card_service, header, trace)


def _send_apdu_T0_case_2s(card_service: _CardService, header: str, Le: str, trace=None):
    """send_apdu_T0_case_2s():
    """
    # Short Le field
    if len(Le) == 2:
        pass
    # Wrong Le field
    else:
        return ['', '', ''], 'Wrong Le value' + Le

    _trace(trace, "Case 2 Command (T=0)")
    [data, SW1, SW2], _ = transmit(card_service, header + Le, trace)
    # Case 2S.1—Process completed: Ne accepted
    if SW1+SW2 == '9000':
        return [data, SW1, SW2], ''
    # Case 2S.2—Process aborted: Ne definitively not accepted
    if SW1+SW2 == '6700':
        return ['', SW1, SW2], 'Error—Process aborted: Ne definitively not accepted (6700)'
    # Case 2S.3—Process aborted; Ne not accepted, Na indicated
    if SW1 == '6C':
        return transmit(card_service, header + SW2, trace)
    # Case 2S.4—SW12 = '9XYZ', except for '9000'
    if SW1.startswith('9'):
        return [data, SW1, SW2], ''

    return [data, SW1, SW2], 'Unknown command case'


def _send_apdu_T0_case_3s(card_service: _CardService, header: str, Lc: str, data: str, trace=None):
    """send_apdu_T0_case_3s():
    """
    # Short Lc field
    if len(Lc) == 2:
        pass
    # Data field too long
    else:
        return ['', '', ''], "Data field too long: " + Lc

    _trace(trace, "Case 3 Command (T=0)")
    return transmit(card_service, header + Lc + data, trace)


def _send_apdu_T0_case_4s(card_service: _CardService, header: str, Lc: str, data: str, Le: str, trace=None):
    """send_apdu_T0_case_4s():
    """
    # Short Lc field
    if len(Lc) == 2:
        pass
    # Data field too long
    else:
        return ['', '', ''], "Data field too long: " + Lc

    # Short Le field
    if len(Le) == 2:
        pass
    # Wrong Le field
    else:
        return ['', '', ''], 'Wrong Le value' + Le

    _trace(trace, "Case 4 Command (T=0)")
    [data, SW1, SW2], _ = transmit(
        card_service, header + Lc + data + Le, trace)
    # Case 4S.1—Process aborted
    if SW1[0] == '6' and SW1[1] in '0456789ABCDEF':
        return [data, SW1, SW2], F'Error—Process aborted ({SW1+SW2})'
    # Case 4S.2—Process completed
    if SW1+SW2 == '9000':
        # Send GET RESPONSE
        return _send_apdu_T0_case_2s(card_service, header[0:2] + 'C00000', Le, trace)
    # Case 4S.3—Process completed with information added
    if SW1 == '61':
        if Le == '00':
            Ne = '100'
        else:
            Ne = Le
        return _send_apdu_T0_case_2s(card_service, header[0:2] + 'C00000', _min(Ne, SW2), trace)
    # Case 4S.4—SW12 = '62XY', or '63XY', or '9XYZ', except for '9000'
    if SW1.startswith('9') or SW1 in ['62', '63']:
        return [data, SW1, SW2], ''

    return [data, SW1, SW2], 'Unknown command case'


def disconnect(card_service: _CardService):
    """disconnect()
    """
    card_service.connection.disconnect()

#
# Helper functions
#


def _trace(trace, message: str):
    if trace is not None:
        trace(message)



"""commands.py
"""

# Standard library imports

# Third party imports

# Local application imports
from common.ber import encode_length as length
from common.ber import encode


def SELECT(aid: str) -> dict:
    """SELECT(): generate APDU for SELECT command
    """
    known_aid_list = {'MASTERCARD' : 'A0000000041010', 'MAESTRO' : 'A00000000410103060', 'VISA' : 'A0000000031010'}
    return {'header' : '00A40400', 'data' : known_aid_list.get(aid, aid), 'Le' : '00'}

def GET_PROCESSING_OPTIONS(pdol: str = '') -> dict:
    """GET_PROCESSING_OPTIONS(): generate APDU for GET PROCESSING OPTIONS command
    """
    return {'header' : '80A80000', 'data' : encode('83', pdol), 'Le' : '00'}

def GPO(pdol: str) -> dict:
    """GPO(): generate APDU for GET PROCESSING OPTIONS command
    """
    return GET_PROCESSING_OPTIONS(pdol)

def READ_RECORD(SFI:int, record:int) -> dict:
    """READ_RECORD(): generate APDU for READ RECORD command
    """
    return {'CLA' : '00', 'INS' : 'B2', 'P1' : F"{record:02X}", 'P2' : F"{SFI*8+4:02X}", 'Le' : '00'}

def GET_DATA(tag: str) -> dict:
    """GET_DATA: generate APDU for GET DATA command
    """
    return {'header' : '80CA' + tag, 'Le' : '00'}


"""str.py

Generic functions working on strings
"""

# Standard library imports

# Third party imports

# Local application imports

def perm(str_in, permutation):
    """perm():
    """
    str_out = ''

    for i in permutation:
        str_out += str_in[i-1]

    return str_out


def exp(str_in, expansion):
    """exp():
    """
    return perm(str_in, expansion)


def lcs(str_in, shift):
    """lcs():
    """
    return str_in[shift:] + str_in[:shift]


def to_ascii(str_in):
    return ''.join([F"{ord(c):02X}" for c in str_in])

def clean (str_in: str, keep_char = '', remove_char = '', command_name='()', str_name='') -> str:
    """clean(str)

    - Removes '_' characters
    - Raises exceptions on illegal characters
    """
    str_out = ''
    for nibble in str_in:
        if nibble in keep_char:
            str_out += nibble
        elif nibble in remove_char:
            pass
        else:
            if str_name == '':
                raise TypeError(F"{command_name}: Unknown nibble: {nibble}")
            else:
                raise TypeError(
                    F"{command_name}: Unknown nibble in {str_name}: {nibble}")

    return str_out


"""records.py
"""

# Standard library imports

# Third party imports

# Local application imports
from common import hstr as _hstr

def decode_afl(AFL:str):
    """decode_afl(): 
    """
    AFL = _hstr.clean(AFL)
    if len(AFL) % 8 != 0:
        # Error if the AFL is not a multiple of 4 bytes
        return None, "AFL is not a multiple of 4 bytes"
    
    afl = []
    for afl_element in _hstr.split_by_length(AFL, 8):
        r, err = _decode_afl_element(afl_element)
        if err is not None:
            return None, err
        else:
            afl.append(r)
    
    return afl, None

#
# Helper functions (assume a clean 'hstr' as input)
#


def _decode_afl_element(e:str):
    (byte1, byte2, byte3, byte4) = _hstr.split_by_length(e, 2)

    if _hstr.and_(byte1, '07') != '00':
        # Error if three least  significant bits are not zeroes
        return None, "AFL Byte 1, bits 3–1 not zeroes"

    SFI = int(byte1, 16)//8

    if byte2 == '00':
        # Error if the second byte is zero
        return None, "AFL Byte 2 is zero"
    first_record = int(byte2, 16)
    
    last_record = int(byte3, 16)
    if last_record < first_record:
        # Error if the last record is smaller than the first record
        return None, "AFL Byte 3 is smaller than Byte 2"
    
    oda_records = int(byte4, 16)
    if oda_records > (last_record - first_record + 1):
        # Error if the number of records used for ODA is greater than the total number of records
        return None, "AFL Byte 4 is greater than total number of records"
    
    return (SFI, first_record, last_record, oda_records), None


"""key_management.py
"""

# Standard library imports

# Third party imports

# Local application imports
from common import hstr
from crypto import des

def master_key_derivation_A(imk, pan, psn = '00'):
    """master_key_derivation_A(): ICC Master Key derivation (Option A)
    """
    X = pan + psn
    if len(X) < 16:
        X = '0' * (16 - len(X)) + X

    Y = X[-16:]

    Zl = des.tdea_2_ede(imk, Y)
    Zr = des.tdea_2_ede(imk, hstr.not_(Y))

    return Zl + Zr



"""intlist.py
"""

# Standard library imports

# Third party imports

# Local application imports


def to_hstr(intlist_in, command_name='()', arg_name=''):
    """to_hstr()
    """
    hstr = ''
    for i in intlist_in:
        if not isinstance(i, int):
            raise TypeError(
                F"{command_name}: wrong element in {arg_name}: {i}")

        hstr += F"{i:02X}"

    return hstr



"""dsc.py: Dynamic Security Codes module
"""

# Standard library imports

# Third party imports

# Local application imports
from common import hstr
from crypto import des

def generate_dcvv(udk: str, pan: str, atc: str, expiration_date: str):
    alternate_pan = F"{atc}{pan[4:]}"

    input_block = F"{alternate_pan}{expiration_date}"
    input_block = input_block + '0' * (32 - len(input_block))

    udk_a = udk[0:16]
    udk_b = udk[16:32]
    block_a = input_block[0:16]
    block_b = input_block[16:32]

    block_c = des.dea_e(udk_a, block_a)
    block_d = hstr.xor(block_c, block_b)
    block_e = des.dea_e(udk_a, block_d)
    block_f = des.dea_d(udk_b, block_e)
    block_g = des.dea_e(udk_a, block_f)
    block_h = hstr.dscan_decimalize(block_g)

    return block_h[0:3]

def generate_ivcvc3(udk: str, track: str) -> str:
    if (len(track) % 16) == 0:
        block = track + "8000000000000000"
    else:
        block = track + "80" + "0" * (16 - len(track) % 16 - 2)

    mac = des.mac_2_ede(udk, block)

    return mac[-4:]

def generate_cvc3(udk: str, block: str) -> str:
    return F"{int(des.tdea_2_ede(udk, block)[-4:], 16):05d}"
